package com.chelvc.framework.base.interceptor;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Map;
import javax.servlet.FilterChain;
import javax.servlet.ReadListener;
import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletRequestWrapper;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.LoggingContextHolder;
import com.chelvc.framework.base.context.Result;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import lombok.extern.slf4j.Slf4j;
import org.springframework.http.HttpStatus;
import org.springframework.web.filter.OncePerRequestFilter;

/**
 * 请求调用拦截器
 *
 * @author Woody
 * @date 2024/4/3
 */
@Slf4j
public class RequestInvokeInterceptor extends OncePerRequestFilter {
    /**
     * 缓存数据流对象
     */
    static class CachingInputStream extends ServletInputStream {
        private final InputStream delegate;

        private CachingInputStream(byte[] bytes) {
            this.delegate = new ByteArrayInputStream(bytes);
        }

        @Override
        public int read() throws IOException {
            return this.delegate.read();
        }

        @Override
        public int read(byte[] b, int off, int len) throws IOException {
            return this.delegate.read(b, off, len);
        }

        @Override
        public int read(byte[] b) throws IOException {
            return this.delegate.read(b);
        }

        @Override
        public long skip(long n) throws IOException {
            return this.delegate.skip(n);
        }

        @Override
        public int available() throws IOException {
            return this.delegate.available();
        }

        @Override
        public void close() throws IOException {
            this.delegate.close();
        }

        @Override
        public synchronized void mark(int limit) {
            this.delegate.mark(limit);
        }

        @Override
        public synchronized void reset() throws IOException {
            this.delegate.reset();
        }

        @Override
        public boolean markSupported() {
            return this.delegate.markSupported();
        }

        @Override
        public boolean isReady() {
            try {
                return this.delegate.available() > 0;
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        @Override
        public boolean isFinished() {
            try {
                return this.delegate.available() == 0;
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        @Override
        public void setReadListener(ReadListener listener) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * 请求缓存包装对象
     */
    static class CachingRequestWrapper extends HttpServletRequestWrapper {
        private final byte[] body;
        private final Map<String, String[]> parameters;

        private CachingRequestWrapper(HttpServletRequest request) {
            super(request);
            if (request instanceof CachingRequestWrapper) {
                this.body = ((CachingRequestWrapper) request).body;
                this.parameters = ((CachingRequestWrapper) request).parameters;
            } else {
                // Servlet 规范规定如果调用了getInputStream()或getReader()方法，则getParameter*()方法的行为是未定义的
                // Tomcat、Jetty、Undertow容器均支持在调用getParameter*()方法后能正常调用getInputStream()方法
                this.parameters = request.getParameterMap();
                this.body = HttpUtils.getBody(request);
            }
        }

        @Override
        public String getParameter(String name) {
            String[] values = this.parameters.get(name);
            return ObjectUtils.isEmpty(values) ? null : values[0];
        }

        @Override
        public Map<String, String[]> getParameterMap() {
            return Collections.unmodifiableMap(this.parameters);
        }

        @Override
        public Enumeration<String> getParameterNames() {
            Iterator<String> iterator = this.parameters.keySet().iterator();
            return new Enumeration<String>() {
                @Override
                public boolean hasMoreElements() {
                    return iterator.hasNext();
                }

                @Override
                public String nextElement() {
                    return iterator.next();
                }
            };
        }

        @Override
        public String[] getParameterValues(String name) {
            return this.parameters.get(name);
        }

        @Override
        public BufferedReader getReader() throws IOException {
            return new BufferedReader(new InputStreamReader(this.getInputStream()));
        }

        @Override
        public ServletInputStream getInputStream() throws IOException {
            return new CachingInputStream(this.body);
        }
    }

    @Override
    protected void doFilterInternal(HttpServletRequest request, HttpServletResponse response, FilterChain chain)
            throws ServletException, IOException {
        if (HttpUtils.isBodyMethod(request) && !HttpUtils.isMultipartBody(request)) {
            request = new CachingRequestWrapper(request);
        }

        try {
            chain.doFilter(request, response);
        } catch (Exception e) {
            LoggingContextHolder.warn(log, request, e);
            String message = ApplicationContextHolder.getMessage("Request.Invalid");
            Result<?> result = Result.of(HttpStatus.BAD_REQUEST.name(), null, message);
            SessionContextHolder.response(response, result);
        }
    }
}
